//----------------------------------------------------------------------------------------
// UI Events protocol
//----------------------------------------------------------------------------------------

#include <Clipbd.xh>								// JWM
#include <ISOStr.h>
#include <Link.xh>
#include <LinkSpec.xh>
#include <LinkSrc.xh>
#include <Undo.xh>

#ifndef _HEADLINE_
#include "CPHeadline.h"
#endif

#include "DartmouthODUtils.h"						// JWM

// JWM
//----------------------------------------------------------------------------------------
// CPHeadline::SetData
//----------------------------------------------------------------------------------------
void CPHeadline::SetData(Environment* ev, char *data, ODChangeID id)
{
		/* generate a change id if we don't have one */
		if (!id)
			id = fSession->UniqueChangeID(ev);
			
		/* remember the change id */
		fChangeID = id;
		
		/* set text data */
		strcpy((char *) fTextData, data);
		CtoPstr((char *) fTextData);
		
		/* mark as dirty */
		this->PrivSetDirty(ev);
		
		/* update link(s) */
		this->UpdateSourceLink(ev, id);
		
		/* redraw */
		this->PrivInvalAllDisplayFrames(ev);
} /* SetData() */

//----------------------------------------------------------------------------------------
// CPHeadline::DoCutOrCopy
//----------------------------------------------------------------------------------------
ODBoolean CPHeadline::DoCutOrCopy(Environment* ev, ODFrame* frame, Boolean cutflag)
{
	EnteringMethod("\pCPHeadline::DoCutOrCopy");

	/* copy data to clipboard */
	ODArbitrator	*arbitrator = fSession->GetArbitrator(ev);
	ODTypeToken		clipboardFocus = fSession->Tokenize(ev, kODClipboardFocus);
	
	if ((frame == arbitrator->GetFocusOwner(ev, clipboardFocus)) ||
		(arbitrator->RequestFocus(ev, clipboardFocus, frame)) )
	{
		/* first clear existing clipboard contents */
		ODClipboard		*volatile clipboard = fSession->GetClipboard(ev);
		
		clipboard->Clear(ev);
		
		/* now put our text on the clipboard */
		ODStorageUnit				*clipstorage = clipboard->GetContentStorageUnit(ev);
		ODDraft						*volatile fromDraft = fSelf->GetStorageUnit(ev)->GetDraft(ev);
		ODDraft						*toDraft = clipstorage->GetDraft(ev);
		ODDraftKey					volatile key = 0;
		
		TRY
		{
			// Always modify the clipboard within a BeginClone-EndClone transaction
			key = fromDraft->BeginClone(ev, toDraft, cutflag ? kODCloneCut : kODCloneCopy);
			
			// Write out the contents, preferred representation first.
			// The type of the best representation is usually the parts own 
			// proprietary format.
			clipstorage->AddProperty(ev, kODPropContents);
			clipstorage->AddValue(ev, kODAppleTEXT);
			ODByteArray	value;
			SetODByteArray(kHeadlineTextMax, fTextData[0], &fTextData[1], &value);
			clipstorage->SetValue(ev, &value);
			
			// if it's a copy, write out a link spec too
			if (!cutflag && (fSrcLink == kODNULL))
			{
				clipstorage->AddProperty(ev, kODPropLinkSpec);
				char	dummy[1];
				SetODByteArray(1, 1, dummy, &value);
				ODLinkSpec* linkSpec = fromDraft->CreateLinkSpec(ev, fSelf, &value);
				linkSpec->WriteLinkSpec(ev, clipstorage);
				delete linkSpec;
			}
			
			// Complete the clone transaction
			fromDraft->EndClone(ev, key);
		}
		CATCH_ALL
		{
			// If an exception was raised, clear the clipboard.
			//   The previous contents of the clipboard are lost.
			if (key != 0)
				fromDraft->AbortClone(ev, key);
			clipboard->Clear(ev);					// invalidates clipstorage
			SysBeep(2);
			// There's no one to catch this error
			// RERAISE;
		}
		ENDTRY
		
		fClipChangeID = clipboard->GetChangeID(ev);
		arbitrator->RelinquishFocus(ev, clipboardFocus, frame);
	} /* with clip focus */
	
	/* if we're cutting, nuke the data */
	if (cutflag)
		this->SetData(ev, "", 0);

	return true;
} /* DoCutOrCopy() */

//----------------------------------------------------------------------------------------
// CPHeadline::MyHandleCutEvent
//----------------------------------------------------------------------------------------
ODBoolean CPHeadline::MyHandleCutEvent(Environment* ev, ODFrame* frame, ODEventData* event)
{
	EnteringMethod("\pCPHeadline::MyHandleCutEvent");

	ODActionData		action;
	DumbPartActionData	my_action;
	
	/* make action record */
	my_action.fActionCode = kODCommandCut;
	my_action.fFrame = frame;
	PtoCstr(fTextData);
	strcpy(my_action.fActionData, (char *) fTextData);
	CtoPstr((char *) fTextData);
	SetODByteArray(sizeof(DumbPartActionData), sizeof(DumbPartActionData), &my_action, &action);
	
	/* add action to undo history */
	ODIText*	undoActionName = CreateIText(smRoman, langEnglish, "Undo Cut");
	ODIText*	redoActionName = CreateIText(smRoman, langEnglish, "Redo Cut");
	fSession->GetUndo(ev)->AddActionToHistory(ev, fSelf,
									&action, kODSingleAction,
									undoActionName, redoActionName);
	DisposeIText(undoActionName);
	DisposeIText(redoActionName);

	return this->DoCutOrCopy(ev, frame, true);
} /* MyHandleCutEvent() */

//----------------------------------------------------------------------------------------
// CPHeadline::MyHandleCopyEvent
//----------------------------------------------------------------------------------------
ODBoolean CPHeadline::MyHandleCopyEvent(Environment* ev, ODFrame* frame, ODEventData* event)
{
	EnteringMethod("\pCPHeadline::MyHandleCopyEvent");

	ODActionData		action;
	DumbPartActionData	my_action;
	
	/* make action record */
	my_action.fActionCode = kODCommandCopy;
	my_action.fFrame = frame;
	PtoCstr(fTextData);
	strcpy(my_action.fActionData, (char *) fTextData);
	CtoPstr((char *) fTextData);
	SetODByteArray(sizeof(DumbPartActionData), sizeof(DumbPartActionData), &my_action, &action);
	
	/* add action to undo history */
	ODIText*	undoActionName = CreateIText(smRoman, langEnglish, "Undo Copy");
	ODIText*	redoActionName = CreateIText(smRoman, langEnglish, "Redo Copy");
	fSession->GetUndo(ev)->AddActionToHistory(ev, fSelf,
									&action, kODSingleAction,
									undoActionName, redoActionName);
	DisposeIText(undoActionName);
	DisposeIText(redoActionName);

	return this->DoCutOrCopy(ev, frame, false);
} /* MyHandleCopyEvent() */

//----------------------------------------------------------------------------------------
// CPHeadline::DoPaste
//----------------------------------------------------------------------------------------
ODBoolean CPHeadline::DoPaste(Environment* ev, ODFrame* frame)
{
	char	pasted[kHeadlineTextMax];
	EnteringMethod("\pCPHeadline::DoPaste");

	/* get data from clipboard */
	ODArbitrator	*arbitrator = fSession->GetArbitrator(ev);
	ODTypeToken		clipboardFocus = fSession->Tokenize(ev, kODClipboardFocus);
	
	if ((frame == arbitrator->GetFocusOwner(ev, clipboardFocus)) ||
		(arbitrator->RequestFocus(ev, clipboardFocus, frame)) )
	{
		ODClipboard		*volatile clipboard = fSession->GetClipboard(ev);
				
		/* get text from the clipboard */
		ODStorageUnit				*clipstorage = clipboard->GetContentStorageUnit(ev);
		ODDraft						*toDraft = fSelf->GetStorageUnit(ev)->GetDraft(ev);
		ODDraft						*volatile fromDraft = clipstorage->GetDraft(ev);
		ODDraftKey					volatile key = 0;
		
		TRY
		{
			// Always modify the clipboard within a BeginClone-EndClone transaction
			key = fromDraft->BeginClone(ev, toDraft, kODClonePaste);
			
			// Read in the contents
			clipstorage->Focus(ev, kODPropContents, kODPosUndefined,   
					kODAppleTEXT, 0, kODPosUndefined);
			ODByteArray	value;
			SetODByteArray(kHeadlineTextMax - 1, kHeadlineTextMax - 1, pasted, &value);
			clipstorage->GetValue(ev, &value);
			pasted[clipstorage->GetSize(ev)] = '\0';

			// Complete the clone transaction
			fromDraft->EndClone(ev, key);
		}
		CATCH_ALL
		{
			// If an exception was raised, clear the clipboard.
			//   The previous contents of the clipboard are lost.
			if (key != 0)
				fromDraft->AbortClone(ev, key);
			SysBeep(2);
			// There's no one to catch this error
			// RERAISE;
		}
		ENDTRY
		
		arbitrator->RelinquishFocus(ev, clipboardFocus, frame);
	} /* with clip focus */
	
	/* use pasted data */
	this->SetData(ev, pasted, 0);

	return true;
} /* DoPaste() */

//----------------------------------------------------------------------------------------
// CPHeadline::MyHandlePasteEvent
//----------------------------------------------------------------------------------------
ODBoolean CPHeadline::MyHandlePasteEvent(Environment* ev, ODFrame* frame, ODEventData* event)
{
	EnteringMethod("\pCPHeadline::MyHandlePasteEvent");

	ODActionData		action;
	DumbPartActionData	my_action;
	
	/* make action record */
	my_action.fActionCode = kODCommandPaste;
	my_action.fFrame = frame;
	PtoCstr(fTextData);
	strcpy(my_action.fActionData, (char *) fTextData);
	CtoPstr((char *) fTextData);
	SetODByteArray(sizeof(DumbPartActionData), sizeof(DumbPartActionData), &my_action, &action);
	
	/* add action to undo history */
	ODIText*	undoActionName = CreateIText(smRoman, langEnglish, "Undo Paste");
	ODIText*	redoActionName = CreateIText(smRoman, langEnglish, "Redo Paste");
	fSession->GetUndo(ev)->AddActionToHistory(ev, fSelf,
									&action, kODSingleAction,
									undoActionName, redoActionName);
	DisposeIText(undoActionName);
	DisposeIText(redoActionName);

	return this->DoPaste(ev, frame);
} /* MyHandlePasteEvent() */

//----------------------------------------------------------------------------------------
// CPHeadline::DoPasteAs
//----------------------------------------------------------------------------------------
ODBoolean CPHeadline::DoPasteAs(Environment* ev, ODFrame* frame, ODPasteAsResult *rslt, ODBoolean dontask)
{
	EnteringMethod("\pCPHeadline::DoPasteAs");
	
	ODArbitrator* arbitrator = fSession->GetArbitrator(ev);
	ODTypeToken clipboardFocus = fSession->Tokenize(ev, kODClipboardFocus);
	
	if ((frame == arbitrator->GetFocusOwner(ev, clipboardFocus)) || 
		(arbitrator->RequestFocus(ev, clipboardFocus, frame)))
	{
		ODClipboard* clipboard = fSession->GetClipboard(ev);
		ODFrameFacetIterator* facets = (ODFrameFacetIterator*) kODNULL;
		ODPasteAsResult pasteAsRslt;
		
		pasteAsRslt.pasteLinkSetting = kODTrue;
		TRY
		{
			ODStorageUnit* clipContentSU = clipboard->GetContentStorageUnit(ev);
			facets = frame->CreateFacetIterator(ev);
			if (dontask)
				pasteAsRslt = *rslt;
			if (dontask || 
				clipboard->ShowPasteAsDialog(ev, 
						kODTrue,
						kODFalse,
						facets->First(ev),
						fSession->Tokenize(ev, kODViewAsFrame),
						&pasteAsRslt))
			{
				// Cannot paste a link if translation is required
				if (pasteAsRslt.translateKind != kODNULL)
					DebugStr("\pcannot translate !!!");

				// Get the root storage unit for the clipboard
				ODStorageUnit* clipRootSU = clipboard->GetContentStorageUnit(ev);
				
				// read text from clipboard
//				clipRootSU->Focus(ev, kODPropContents, kODPosUndefined,   
//						kODAppleTEXT, 0, kODPosUndefined);
//				ODByteArray	value;
//				char		pasted[kHeadlineTextMax];
//				SetODByteArray(kHeadlineTextMax - 1, kHeadlineTextMax - 1, pasted, &value);
//				clipRootSU->GetValue(ev, &value);
//				pasted[clipRootSU->GetSize(ev)] = '\0';
//				this->SetData(ev, pasted, 0);

				// read link from clipboard
				if (clipRootSU->Exists(ev, kODPropLinkSpec, (ODValueType) kODNULL, 0)) 
				{
					ODDraft		*draft = fStorageUnit->GetDraft(ev);
					ODLinkSpec	*linkSpec = draft->CreateLinkSpec(ev, fSelf, 0);
					clipRootSU->Focus(ev, kODPropLinkSpec, kODPosUndefined, (ODValueType)kODNULL, 0, kODPosUndefined);
					linkSpec->ReadLinkSpec(ev, clipRootSU);
					TRY
					{
						// get rid of existing link
						if (fDestLink != kODNULL)
							this->RemoveDestLink(ev);
							
						fDestLink = draft->GetLink(ev, (ODStorageUnitID) kODNULL, linkSpec);
						if (fDestLink != (ODLink*) kODNULL)
						{
							fDestLinkInfo.change = fDestLink->GetChangeID(ev);
							GetDateTime((ODTime*) &fDestLinkInfo.creationTime);
							fDestLinkInfo.changeTime = fDestLinkInfo.creationTime;
							fDestLinkInfo.kind = pasteAsRslt.selectedKind;
							fDestLinkInfo.autoUpdate = pasteAsRslt.autoUpdateSetting;
				
							// Now register as a dependent of the link
							// Specify kODUnknownChange to force getting the text.
							// By default the destination is automatically updated.
							if (fDestLinkInfo.autoUpdate)
								fDestLink->RegisterDependent(ev, fSelf, kODUnknownChange);
							draft->SetChangedFromPrev(ev);
							this->LinkUpdated(ev, fDestLink, fDestLink->GetChangeID(ev));
						}
						else 
						{
							SysBeep(30);
							DebugStr("\pCould not GetLink");
						}
					}
					CATCH(kODErrDocNotSaved)
						DebugStr("\pSource document not saved!");
					CATCH_ALL
					ENDTRY
					delete linkSpec;
				} /* clipRootSU->Exists(linkspec) */
				else
				{
//					DebugStr("\pCould not find link spec on clipboard");
					pasteAsRslt.pasteLinkSetting = kODFalse;
				}
				
				arbitrator->RelinquishFocus(ev, clipboardFocus, frame);	
			}  /* if ShowPasteAsDialog() */
		}
		CATCH_ALL
		{
			somPrintf("DoPasteAs caught error\n");
		}
		ENDTRY
		delete facets;
		
		*rslt = pasteAsRslt;
		return pasteAsRslt.pasteLinkSetting == kODTrue;
	}
	
	return false;
} /* DoPasteAs() */

//----------------------------------------------------------------------------------------
// CPHeadline::MyHandlePasteAsEvent
//----------------------------------------------------------------------------------------
ODBoolean CPHeadline::MyHandlePasteAsEvent(Environment* ev, ODFrame* frame, ODEventData* event)
{
	EnteringMethod("\pCPHeadline::MyHandlePasteAsEvent");
	
	// get set to undo
	ODActionData		action;
	DumbPartActionData	my_action;
	ODPasteAsResult 	pasteAsRslt;

	/* start making action record */
	my_action.fActionCode = kODCommandPasteAs;
	my_action.fFrame = frame;
	PtoCstr(fTextData);
	strcpy(my_action.fActionData, (char *) fTextData);
	CtoPstr((char *) fTextData);
	
	ODBoolean result = this->DoPasteAs(ev, frame, &pasteAsRslt, false);
	
	if (result)
	{
		/* add action to undo history */
		my_action.fPasteAsRslt = pasteAsRslt;
		SetODByteArray(sizeof(DumbPartActionData), sizeof(DumbPartActionData), &my_action, &action);

		ODIText*	undoActionName = CreateIText(smRoman, langEnglish, "Undo Paste As");
		ODIText*	redoActionName = CreateIText(smRoman, langEnglish, "Redo Paste As");
		fSession->GetUndo(ev)->AddActionToHistory(ev, fSelf,
										&action, kODSingleAction,
										undoActionName, redoActionName);
		DisposeIText(undoActionName);
		DisposeIText(redoActionName);
	}
	
	return result;
}
				
			
//----------------------------------------------------------------------------------------
// CPHeadline::DoClear
//----------------------------------------------------------------------------------------
ODBoolean CPHeadline::DoClear(Environment* ev)
{
	EnteringMethod("\pCPHeadline::DoClear");

	/* nuke the data */
	this->SetData(ev, "", 0);
	
	return true;
} /* DoClear() */

//----------------------------------------------------------------------------------------
// CPHeadline::MyHandleClearEvent
//----------------------------------------------------------------------------------------
ODBoolean CPHeadline::MyHandleClearEvent(Environment* ev, ODFrame* frame, ODEventData* event)
{
	EnteringMethod("\pCPHeadline::MyHandleClearEvent");
ODUnused(frame);

	ODActionData		action;
	DumbPartActionData	my_action;
	
	/* make action record */
	my_action.fActionCode = kODCommandClear;
	my_action.fFrame = frame;
	PtoCstr(fTextData);
	strcpy(my_action.fActionData, (char *) fTextData);
	CtoPstr((char *) fTextData);
	SetODByteArray(sizeof(DumbPartActionData), sizeof(DumbPartActionData), &my_action, &action);
	
	/* add action to undo history */
	ODIText*	undoActionName = CreateIText(smRoman, langEnglish, "Undo Clear");
	ODIText*	redoActionName = CreateIText(smRoman, langEnglish, "Redo Clear");
	fSession->GetUndo(ev)->AddActionToHistory(ev, fSelf,
									&action, kODSingleAction,
									undoActionName, redoActionName);
	DisposeIText(undoActionName);
	DisposeIText(redoActionName);

	return this->DoClear(ev);
} /* MyHandleClearEvent() */

//----------------------------------------------------------------------------------------
// CPHeadline::MyHandleGetPartInfo
//----------------------------------------------------------------------------------------
ODBoolean CPHeadline::MyHandleGetPartInfo(Environment* ev, ODFrame* frame, ODEventData* event)
{
	ODLinkInfoResult infoResult;
	ODFrameFacetIterator *facets = frame->CreateFacetIterator(ev);

	if ((fSrcLink != kODNULL) && !(event->modifiers & optionKey))
	{
		if (fSrcLink->ShowLinkSourceInfo(ev, 
						facets->First(ev),
						fSrcLink->GetChangeID(ev), 
						kODTrue,
						&infoResult))
		{
			switch (infoResult.action)
			{
				case kODLinkInfoBreakLink:
					RemoveSourceLink(ev);
					break;
				case kODLinkInfoUpdateNow:
					UpdateSourceLink(ev, fChangeID);
					break;
				case kODLinkInfoOk :
					if (infoResult.autoUpdate != fSrcLink->IsAutoUpdate(ev))
					{
						fSrcLink->SetAutoUpdate(ev, infoResult.autoUpdate);
						if (infoResult.autoUpdate)
								// Changed from manual to automatic updates, and the link
								// is out of date; update it now
								UpdateSourceLink(ev, fChangeID);
					}
					break;
			}
		}
	}
	else if ((fDestLink != kODNULL) && !(event->modifiers & optionKey))
	{
		if (fDestLink->ShowLinkDestinationInfo(ev, 
										facets->First(ev),
										&fDestLinkInfo, 
										kODTrue,
										&infoResult))
		{
			switch (infoResult.action)
			{
			case kODLinkInfoFindSource:
				TRY
					fDestLink->ShowSourceContent(ev);
				CATCH_ALL
//					DebugStr("\pDumbPart : Error returned by ODLink::ShowSourceContent" );
					::SysBeep(2);
				ENDTRY;
				break;
			case kODLinkInfoBreakLink:
				RemoveDestLink(ev);
				break;
			case kODLinkInfoUpdateNow:
				LinkUpdated(ev, fDestLink, fDestLink->GetChangeID(ev));
				break;
			case kODLinkInfoOk:
				if (infoResult.autoUpdate != fDestLinkInfo.autoUpdate)
				{
					fDestLinkInfo.autoUpdate = infoResult.autoUpdate;
					if (infoResult.autoUpdate)
						fDestLink->RegisterDependent(ev, fSelf, fDestLinkInfo.change);
					else
						fDestLink->UnregisterDependent( ev, fSelf);
				}
				break;
			default:
				break;
			}
		}
	}
	else
		fSession->GetInfo(ev)->ShowPartFrameInfo(ev, facets->First(ev), kODTrue);
	
	delete facets;
	
	return true;
} /* MyHandleGetPartInfo() */

//----------------------------------------------------------------------------------------
// CPHeadline::DoFontChange
//----------------------------------------------------------------------------------------
ODBoolean CPHeadline::DoFontChange(Environment* ev, StringPtr fontname)
{
	EnteringMethod("\pCPHeadline::DoFontChange");

	GetFNum(fontname, &fFontNo);
	
	if (fFontNo == 0)
	{
		Str255	hold;
		
		GetFontName(0, hold);
		if (!EqualString(fontname, hold, false, true))
			/* this font wasn't found, use application font */
			fFontNo = applFont;
	}

	/* mark as dirty */
	this->PrivSetDirty(ev);
	
	/* redraw */
	this->PrivInvalAllDisplayFrames(ev);

	return true;
} /* DoFontChange() */

//----------------------------------------------------------------------------------------
// CPHeadline::MyHandleChangeFontEvent
//----------------------------------------------------------------------------------------
ODBoolean CPHeadline::MyHandleChangeFontEvent(Environment* ev, ODFrame* frame, StringPtr fontname)
{
	EnteringMethod("\pCPHeadline::MyHandleChangeFontEvent");

	ODActionData		action;
	DumbPartActionData	my_action;
	
	/* make action record w/ old font number and new font name */
	my_action.fActionCode = kHeadlineCommandChangeFont;
	my_action.fFrame = frame;
	my_action.fPrevFont = fFontNo;
	(void) memcpy(my_action.fActionData, fontname, fontname[0] + 1);
	SetODByteArray(sizeof(DumbPartActionData), sizeof(DumbPartActionData), &my_action, &action);
	
	/* add action to undo history */
	ODIText*	undoActionName = CreateIText(smRoman, langEnglish, "Undo Font Change");
	ODIText*	redoActionName = CreateIText(smRoman, langEnglish, "Redo Font Change");
	fSession->GetUndo(ev)->AddActionToHistory(ev, fSelf,
									&action, kODSingleAction,
									undoActionName, redoActionName);
	DisposeIText(undoActionName);
	DisposeIText(redoActionName);

	return this->DoFontChange(ev, fontname);
} /* MyHandleChangeFontEvent() */

// !JWM

//----------------------------------------------------------------------------------------

ODBoolean CPHeadline::HandleEvent(Environment* ev, ODEventData* event,
								   ODFrame* frame, ODFacet* facet,
								   ODEventInfo* eventInfo)
{
	// CodeWarrior: Cmd-Click or Search:Find Definition  Explain_HandleEvent
	EnteringMethod("\pCPHeadline::HandleEvent");
	ODUnused(eventInfo);

	ODBoolean	tWasHandled = kODFalse;

	switch ( event->what )
	{
	case nullEvent:
		if (::IsDialogEvent((EventRecord *)event))
		{
			short itemHit;
			DialogPtr dialog;
			tWasHandled = ::DialogSelect((EventRecord *)event, &dialog, &itemHit);
			return tWasHandled;
		}
		tWasHandled = kODTrue;
		break;

	case mouseDown:
		tWasHandled = this->PrivHandleMouseDown(ev, event, frame, facet);
		break;

	case kODEvtMouseDownEmbedded:
	case mouseUp:
	case keyDown:
	case autoKey:
	case activateEvt:
		tWasHandled = kODTrue;
		break;
		
	case kODEvtMenu:
		tWasHandled = this->PrivHandleMenuEvent(ev, frame, event );
		break;
		
	default:
		break;
	}

	return tWasHandled;
}

//----------------------------------------------------------------------------------------

void CPHeadline::AdjustMenus(Environment* ev, ODFrame* frame)
{
	// CodeWarrior: Cmd-Click or Search:Find Definition  Explain_AdjustMenus
	EnteringMethod("\pCPHeadline::AdjustMenus");
	ODUnused(frame);

	Str63		aboutText;
	ODIText*	odiText;

	fMenuBar->EnableCommand(ev, kODCommandViewAsWin, !frame->IsRoot(ev) );
	fMenuBar->EnableCommand(ev, kODCommandGetPartInfo, kODTrue);

	{
		CUsingLibraryResources fil;
		::GetIndString(aboutText, kContentStringResID, kAboutTextID);
	}

	ODScriptCode script = ::GetScriptManagerVariable(smSysScript);
	ODLangCode lang = ::GetScriptVariable(script, smScriptLang);
	odiText = CreateIText(script, lang, aboutText);
	fMenuBar->SetItemString(ev, kODCommandAbout, odiText);
	DisposeIText(odiText);
	
	// JWM
	/* enable Cut/Copy/Clear item if there's any data */
	fMenuBar->EnableCommand(ev, kODCommandCut, fWritable && !fDestLink && fTextData[0]);
	fMenuBar->EnableCommand(ev, kODCommandCopy, fTextData[0]);
	fMenuBar->EnableCommand(ev, kODCommandClear, fWritable && !fDestLink && fTextData[0]);
	
	/* enable/disable Paste and PasteAs */
	ODArbitrator	*arbitrator = fSession->GetArbitrator(ev);
	ODTypeToken		clipboardFocus = fSession->Tokenize(ev, kODClipboardFocus);
	
	if ((frame == arbitrator->GetFocusOwner(ev, clipboardFocus)) ||
		(arbitrator->RequestFocus(ev, clipboardFocus, frame)))
	{
		ODClipboard		*clipboard = fSession->GetClipboard(ev);
		ODStorageUnit	*clipstorage = clipboard->GetContentStorageUnit(ev);
		
		ODBoolean	canPaste = kODFalse;
		TRY
		{
			long	len;
			
			clipstorage->Focus(ev, kODPropContents, kODPosUndefined,   
					kODAppleTEXT, 0, kODPosUndefined);
			
			len = clipstorage->GetSize(ev);
			canPaste = fWritable && !fDestLink && (len > 0) && (len < kHeadlineTextMax);
			fMenuBar->EnableCommand(ev, kODCommandPaste, canPaste);
		}
		CATCH_ALL
		{
			fMenuBar->EnableCommand(ev, kODCommandPaste, kODFalse);
		}
		ENDTRY
		
		if (canPaste &&
			clipstorage->Exists(ev,kODPropLinkSpec,(ODValueType)kODNULL,0) &&
			(clipboard->GetChangeID(ev) != fClipChangeID))
			fMenuBar->EnableCommand(ev, kODCommandPasteAs, kODTrue);
		else fMenuBar->EnableCommand(ev, kODCommandPasteAs, kODFalse);

		arbitrator->RelinquishFocus(ev, clipboardFocus, frame);
	} /* with clip focus */
	
	// Toggle between "Part Info" and "Link Info" depending on OptionKey
#define shiftKeyItem	1
#define shiftKeyMask	1
#define optionKeyItem	1
#define optionKeyMask	4
#define commandKeyItem	1
#define commandKeyMask	0x8000
#define controlKeyItem	1
#define controlKeyMask	8
	KeyMap theKeyMap;
	GetKeys( theKeyMap );
	if( ((fSrcLink != kODNULL) || (fDestLink != kODNULL)) && 
		!(theKeyMap[optionKeyItem] & optionKeyMask))
		fMenuBar->SetItemString(ev, kODCommandGetPartInfo, PStrToIntl("\pLink Info", kODNULL));
	else
		fMenuBar->SetItemString(ev, kODCommandGetPartInfo, PStrToIntl("\pPart Info", kODNULL));
		
	// fix font menu
	MenuHandle	mh = GetMenuHandle(kMenuID);
	short		index = CountMItems(mh);
	Str255		fontname;
	GetFontName(fFontNo, fontname);
	while (index > 0)
	{
		Str255	hold;
		GetMenuItemText(mh, index, hold);
		CheckItem(mh, index, EqualString(fontname, hold, false, true));
		index--;
	}
	// !JWM
}

//----------------------------------------------------------------------------------------

ODSession* gSession;

ODBoolean CPHeadline::PrivHandleMenuEvent(Environment* ev, ODFrame* frame, ODEventData* event)
{
	EnteringMethod("\pCPHeadline::PrivHandleMenuEvent");
	ODUnused(frame);

	long		menuResult = event->message;
	short		menu = HiWord(menuResult);
	short		item = LoWord(menuResult);
	ODCommandID	command = fMenuBar->GetCommand(ev, menu, item);	// JWM
	
	if (menu) {
		if (fMenuBar->IsCommandSynthetic(ev, command))// Font etc. JWM
		{
			// JWM
			switch (menu)
			{
			case kMenuID:
				Str255	fontname;
				
				GetMenuItemText(GetMenuHandle(kMenuID), item, fontname);
				return this->MyHandleChangeFontEvent(ev, frame, fontname);
				break;
				
			default:
				return kODFalse;
			}
		}
		else switch (command) 
		{
		case kODCommandAbout:
		{
			CUsingLibraryResources fil;

			if (fSession->GetArbitrator(ev)->RequestFocus(ev, fModalFocus, frame))
			{
				::SetCursor(&ODQDGlobals.arrow);
				gSession = fSession;				// DialogFilter uses this!
				ModalFilterUPP modalFilter = NewModalFilterProc(MyDialogFilter);
				::Alert(kHeadlineAboutBoxID, modalFilter);
				DisposeRoutineDescriptor(modalFilter);
				fSession->GetArbitrator(ev)->RelinquishFocus(ev, fModalFocus,frame);
			}
			else
				::SysBeep(1);
		}
		break;

		// JWM
		case kODCommandCut:
			return this->MyHandleCutEvent(ev, frame, event);
			break;

		case kODCommandCopy:
			return this->MyHandleCopyEvent(ev, frame, event);
			break;

		case kODCommandPasteAs :
			// tricky --> may fall through to regular paste
			if (this->MyHandlePasteAsEvent(ev, frame, event))
				return true;
			
		case kODCommandPaste:
			return this->MyHandlePasteEvent(ev, frame, event);
			break;

		case kODCommandClear:
			return this->MyHandleClearEvent(ev, frame, event);
			break;
		// !JWM
		
		case kODCommandViewAsWin:
			this->Open(ev, frame);
			break;

		case kODCommandGetPartInfo:
			// JWM
			return this->MyHandleGetPartInfo(ev, frame, event);
//			ODFrameFacetIterator* facets = frame->CreateFacetIterator(ev);
//			fSession->GetInfo(ev)->ShowPartFrameInfo(ev, facets->First(ev), kODFalse);
			break;

		default:
			return kODFalse;
		}
	}

	return kODTrue;
}

//------------------------------------------------------------------------------

ODBoolean CPHeadline::PrivHandleMouseDown(Environment* ev, ODEventData* event,
										 ODFrame* frame, ODFacet* facet)
{
	EnteringMethod("\pCPHeadline::PrivHandleMouseDown");

	ODBoolean handled = kODFalse;

	if (facet == kODNULL)		// Did we get a click outside a modal dialog? 
	{
		::SysBeep(1);
		return kODTrue;			// Return that click was handled.
	}

	// Activate the window if it is not active
	if (! facet->GetWindow(ev)->IsActive(ev))
	{
		// select the window
		facet->GetWindow(ev)->Select(ev);

		// typically you continue handling the event, if not, just return kODTrue
		handled = kODTrue;
		
		// return kODTrue; // user needs to click twice to get something done
	}

	// dialog events first
	if (::IsDialogEvent((EventRecord *)event))
	{
		short		itemHit;
		DialogPtr	dialog;

		handled = ::DialogSelect((EventRecord *)event, &dialog, &itemHit);
		ODWindow* window = fSession->GetWindowState(ev)->GetODWindow(ev, dialog);
		return handled;		// In preparation for more "if" cases to be added below.
	}

	// get the mouse in local coordinates
	ODWindow* odwnd = frame->GetWindow(ev);
	ODPoint odPoint;
	GetWindowPoint(odwnd, ev, event->where, &odPoint);

	// test for click in grow box, if appropriate
	if (frame->IsRoot(ev))
	{
		WindowPtr wnd   = (WindowPtr)odwnd->GetPlatformWindow(ev);

		Rect r = wnd->portRect;
		r.left = r.right  - 15;
		r.top  = r.bottom - 15;
		if (::PtInRect(odPoint.AsQDPoint(), &r)) {
			::SetRect(&r, 200, 200, 20000, 20000);
			long newWindowSize = ::GrowWindow(wnd, event->where, &r);
			::SizeWindow(wnd, LoWord(newWindowSize), HiWord(newWindowSize), true);
			odwnd->AdjustWindowShape(ev);
			return kODTrue;
		}
	}

	if ( fSession->GetArbitrator(ev)->RequestFocusSet(ev, fFocusSet,frame) )
	{
		// Try to grab all of our foci
		this->FocusAcquired(ev, fSelectionFocus, frame);
		this->FocusAcquired(ev, fMenuFocus, frame);
		this->FocusAcquired(ev, fKeyFocus, frame);

	} else {
		// We couldn't get focus we requested, so bail.
		// Alternatively, parts of functionality can be disabled.
		return kODFalse;
	}

	// Handle clicks in used shape
	odPoint = facet->GetWindowContentTransform(ev, kODNULL)->InvertPoint(ev, &odPoint);

	ODShape*  usedShape = this->PrivGetUsedShape(ev, frame);
	ODBoolean hit       = usedShape->ContainsPoint(ev, &odPoint);
	usedShape->Release(ev);

	if (hit)
		handled = this->PrivUsedShapeHandleMouseDown(ev, event, frame, facet);

	return handled;
}

//------------------------------------------------------------------------------

ODBoolean CPHeadline::PrivUsedShapeHandleMouseDown(Environment* ev,
		ODEventData* event, ODFrame* frame, ODFacet* facet)
{
	EnteringMethod("\pCPHeadline::PrivUsedShapeHandleMouseDown");

	// Get the window and platform window (i.e. WindowPtr)
	ODWindow* odwnd = frame->GetWindow(ev);
	ODPlatformWindow facetPort = odwnd->GetPlatformWindow(ev);

	// Get the mouse in window-local and facet-local coordinates
	ODPoint windowPoint;
	GetWindowPoint(odwnd, ev, event->where, &windowPoint);
	ODPoint facetPoint = 
		facet->GetWindowContentTransform(ev, kODNULL)->InvertPoint(ev, &windowPoint);

	// Handle the click in whatever way is appropriate
	{
		CFocus  foc(ev, facet);
		Point	macPoint = facetPoint.AsQDPoint();
		// You now have the click point in fact coordinate space, so do something.
	}

	return kODTrue;		// We handled the click.
}